package parMSpindle;
/* ParMSpindle... main program control */

/*  <ParMSpindle - a simulation of ParM based plasmid segregation>
    Copyright (C) <2008>  <Jonathan B. Alberts>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.text.DecimalFormat;

import javax.swing.Timer;

//import FileOps.SourceFileFilter;

public class ParMSpindle {
	
	// graphics component
	static ParMSpindle_Graphics parMGraphics;
	static TimeLoop timeLoop;
	
	// output formats
	static DecimalFormat deflectionFormat = new DecimalFormat ("0.0000");
	static DecimalFormat expFormat = new DecimalFormat ("0.000E0");
	
	
	public ParMSpindle (String[] args) {
		parseArgs(args);
		
		if (Env.paramFile != null) { FileOps.loadParamConfig(Env.paramFile, false); }
		if (Env.logFiles) { FileOps.remoteParamConfigSave(); }
		
		if (!Env.remote) { 
			parMGraphics = new ParMSpindle_Graphics( );
			parMGraphics.initialize( );
			parMGraphics.buildUniverse( );
			parMGraphics.showFrame( ); 
		}
		
		// reset dependent parameters, etc
		Env.setTimeStepCounts();
		Env.resetElasticityTstValues(); 
		Env.resetPersistenceTstValues(); 
		
		// make Things
		Bug.makeInitialBug();
		makeInitialThings();
		
		// make the time loop instance
		timeLoop = new TimeLoop();
		timeLoop.start();
	}
	
	protected void parseArgs( String[] args ) {
		for ( int i = 0; i < args.length; i++ ) {
			if (args[i].equals("-help")) {
				talkln (" Command line options for ParMSpindle:");
				talkln (" -help: print this message");
				talkln (" -r  : run the code 'remotely', without any graphics output");
				talkln (" -pf -paramfile -paramFile (file) : load parameter values from specified file");
				talkln (" -ic -initial condition (file) : load initial condition from specified qk file");
				talkln (" -o -out (directory) : create specified directory... all log, histogram, and qk files will be saved there");
				talkln (" -outMade (directory) : same as -out except main directory already made... all log, histogram, and qk files will be saved there");
				talkln (" -lf -logfile -logFile (directory) : create specified directory and save all output files to there");
				talkln (" -qk (directory) : create specified directory and save qk files there.. allows rendering of simulation");
				talkln (" -qkN (int) : number of integration steps between qk file saves... default (" + String.valueOf(Env.toQKFileInterval) + ")");
				talkln (" -biochem : run the simulation without any collisions, forces, or brownian motion");
				talkln (" -oc : ordered filaments (in a biochem only run) are centered");
				System.exit(0);
			}
			
			if ( args[i].equals( "-r" )) {
				Env.remote = true;
				Env.paused = false;
			}
			
			if ((args[i].equals("-paramfile")) | (args[i].equals("-pf")) | (args[i].equals("-paramFile"))) {
				File paramFile = new File(args[i+1]);
				if (paramFile.exists()) {
					Env.paramFile = paramFile;
				} else {
					talkln ("Specified param file can't be found... check path, etc");
					System.exit(0);
				}
			}
			
			if (args[i].equals("-ic")) {
				File icFile = new File(args[i+1]);
				if (icFile.exists()) {
					Env.fromQKStatePath = icFile.getAbsolutePath();
				} else {
					talkln ("Specified ic file can't be found... check path, etc");
					System.exit(0);
				}
			}
			
			
			if (args[i].equals("-biochem")) {
				Env.biochemOnly.setActive(true);
			}
			
			if (args[i].equals("-oc")) {
				Env.orderedCentered = true;
			}
			
			if ((args[i].equals("-o")) | (args[i].equals("-out")) | (args[i].equals("-outfile"))) {
				File outFolder = new File(args[i+1]);
				File altOutFolder = new File (args[i+1]);
				int j=1;
				while (altOutFolder.isDirectory()) {
					System.out.println (altOutFolder.getName() + " exists.... keeping file name BUT changing directory name");
					altOutFolder = new File(args[i+1] + "." + String.valueOf(j));
					j++;
				}
				Env.outFileName = outFolder.getName();
				altOutFolder.mkdir();
				// directory for the log files
				File logSubFolder = new File(altOutFolder.getAbsolutePath() + File.separator + Env.outFileName + "-LOG");
				Env.logFolderPath = logSubFolder.getAbsolutePath();
				logSubFolder.mkdir();
				Env.logFiles = true;
				// directory for the source files
				File srcSubFolder = new File (altOutFolder.getAbsolutePath() + File.separator + Env.outFileName + "-SRC");
				Env.srcFilePath = srcSubFolder.getAbsolutePath();
				srcSubFolder.mkdir();
				// directory for the qk files
				File qkSubFolder = new File(altOutFolder.getAbsolutePath() + File.separator + Env.outFileName + "-QK");
				Env.qkFilePath = qkSubFolder.getAbsolutePath();
				qkSubFolder.mkdir();
				Env.toQKFile = true;
				//*******
				
			}
			
			if (args[i].equals("-outMade")) {
				File outFolder = new File(args[i+1]);
				Env.outFileName = outFolder.getName();
	
				// directory for the log files
				File logSubFolder = new File(outFolder.getAbsolutePath() + File.separator + Env.outFileName + "-LOG");
				Env.logFolderPath = logSubFolder.getAbsolutePath();
				logSubFolder.mkdir();
				Env.logFiles = true;
				// directory for the qk files
				File qkSubFolder = new File(outFolder.getAbsolutePath() + File.separator + Env.outFileName + "-QK");
				Env.qkFilePath = qkSubFolder.getAbsolutePath();
				qkSubFolder.mkdir();
				Env.toQKFile = true;
				//*******
				
			}
			
			if ((args[i].equals("-logfile")) | (args[i].equals("-lf")) | (args[i].equals("-logFile"))) {
				File outFolder = new File(args[i+1]);
				File altOutFolder = new File (args[i+1]);
				int j=1;
				while (altOutFolder.isDirectory()) {
					System.out.println (altOutFolder.getName() + " exists.... keeping file name BUT changing directory name");
					altOutFolder = new File(args[i+1] + "." + String.valueOf(j));
					j++;
				}
				altOutFolder.mkdir();
				Env.outFileName = outFolder.getName();
				Env.logFolderPath = altOutFolder.getAbsolutePath();
				// make a directory for the source files
				File srcdir = new File (altOutFolder.getAbsolutePath() + File.separator + Env.outFileName + ".SRC");
				//srcdir.mkdir();
				Env.srcFilePath = srcdir.getAbsolutePath();
				//*******
				Env.logFiles = true;
			}
			
			if (args[i].equals("-qk")) {
				File qkFolder = new File(args[i+1]);
				File altqkFolder = new File (args[i+1]);
				int j=1;
				while (altqkFolder.isDirectory()) {
					System.out.println (altqkFolder.getName() + " exists.... keeping file name BUT changing directory name");
					altqkFolder = new File(args[i+1] + "." + String.valueOf(j));
					j++;
				}
				altqkFolder.mkdir();
				Env.outFileName = qkFolder.getName();
				Env.qkFilePath = altqkFolder.getAbsolutePath();
				Env.toQKFile = true;
				System.out.println("QK file output to " + Env.qkFilePath);
			}
			
			if (args[i].equals("-qkN")) {
				try {
					Double qkN = Double.valueOf(args[i+1]);
					Env.toQKFileInterval.setValue(qkN.intValue());
					System.out.println("QK file interval set to " + Env.toQKFileInterval.getStringValue());
				} catch (java.lang.NumberFormatException nfe) {
					System.out.println("Invalid qk file interval specified... must be an integer... the setting was ignored");
				}
			}
		}
	}
	
	class TimeLoop extends Thread {
	
		public void run() {
			if (!Env.remote) {
				//parMGraphics.setTopLabels();
				parMGraphics.updateBugScene();		// draw once at the beginning of the simulation
				startUpdateTimer();
			}
			doLoop();
			System.exit(0);
		}
	}
	
	public static void doLoop() {
		boolean paintedThisStep;
		double lastReportTime = System.currentTimeMillis();
		int drawCounter = 0;
		int toFileCounter = 0;
		int remoteOutCounter = 0;
		int qkCounter = 0;
		int collisionCkCounter = 0;
		int ckElasticityCounter = 0;
		int ckPersistenceCounter = 0;
		int applyBrownianForcesCounter = 0;
		int biochemOnlyActionsCounter = 0;
		if (Env.toQKFile) { qkCounter = Env.toQKFileInterval.getIntValue(); } 	// will write initial QK file
		
		while (Env.simulationTime <= Env.runTime.getValue() | Env.fromQKFile) {
			if (Env.paused) { 
				if (Env.viewRotation) {
					parMGraphics.rotateView();  // let's us set up a good looking rotation before starting movie
					try { Thread.sleep(20); } catch (InterruptedException e) { talkln ("error sleeping"); }
				} else {
					try { Thread.sleep(1000); } catch (InterruptedException e) { talkln ("error sleeping"); }
				}
			}
			if (!Env.paused) {
				Env.fromQKFile = false;	// not rendering if in this loop
				synchronized(Env.safeO) {
					 // Plasmid and Segment Collisions
					 if (!Env.biochemOnly.isActive() & (collisionCkCounter >= Thing.collisionCheckInt | Env.simulationTime == 0)) { 
						 Plasmid.checkPlasmidCollision();
						 FilSegment.checkFilSegCollisions();
						 collisionCkCounter = 0;
					 }
					 // Brownian Motion
					 if (!Env.biochemOnly.isActive() & (applyBrownianForcesCounter >= Thing.brownianApplyInt | Env.simulationTime == 0)) {
						 Thing.brownianMotionForAll();
						 applyBrownianForcesCounter = 0;
					 }
					 FilLink.enforceAllFilLinks();
					 for (int i=0;i<Thing.thingCt;i++) {
						Thing.theThings[i].step();
					 }
					
					//update counters and flags
					paintedThisStep = false;
					Env.counter++;
					Env.simulationTime += Env.deltaT.getValue();
					remoteOutCounter++;
					qkCounter++;
					collisionCkCounter++;
					ckElasticityCounter++;
					ckPersistenceCounter++;
					applyBrownianForcesCounter++;
					biochemOnlyActionsCounter++;
					
					if (!Env.remote) { 
						//parMGraphics.setTopLabels();
						drawCounter++;
						toFileCounter++;
						
						if ((Env.paintOn) & (drawCounter >= Env.drawInterval.getIntValue() | Env.simulationTime == 0)) {
							if ((Env.viewRotation) & (!Env.toFile)) { parMGraphics.rotateView(); }
							parMGraphics.updateBugScene();
							paintedThisStep = true;
							drawCounter = 0;
							if (Env.makeFilLengthHisto.isActive()) { FilHistoControl.setFilLengthHisto(); }
							//try { Thread.sleep(20); } catch (InterruptedException e) { talkln ("error sleeping"); }
						}
						
						if ((Env.toFile) & (toFileCounter >= Env.toFileInterval.getIntValue() | parMGraphics.theCanvas.firstRender())) {
							if (!paintedThisStep) {
								if (Env.viewRotation) { parMGraphics.rotateView(); }
								parMGraphics.updateBugScene();
								paintedThisStep = true;
								if (Env.makeFilLengthHisto.isActive()) { FilHistoControl.setFilLengthHisto(); }
								try { Thread.sleep(500); } catch (InterruptedException e) { talkln ("error sleeping"); }
							}
							parMGraphics.captureImage(parMGraphics.theCanvas);
							parMGraphics.waitOnCapture(parMGraphics.theCanvas);
							toFileCounter = 0;
							try { Thread.sleep(20); } catch (InterruptedException e) { talkln ("error sleeping"); }
						}
						
						if ((Env.toQKFile) & (qkCounter >= Env.toQKFileInterval.getIntValue())) {
							FilSegment.updateAllMonomerPositions();
							FileOps.takeQuickPicture(Env.qkFilePath + File.separator + FileOps.makeNameFromIntStep(Env.outFileName,Env.counter) + ".qk");
							qkCounter = 0;
						}
						
						if (Env.cantileverElasticityTst & (ckElasticityCounter >= Env.ckElasticityInterval)) {
							elasticityCk();
							ckElasticityCounter = 0;
						}
						
						if (Env.testingPersistenceLength & (ckPersistenceCounter >= Env.ckPersistenceInterval)) {
							FilSegment.persistenceTrack();
							FilSegment.transDiffTrack();
							ckPersistenceCounter = 0;
						}
						
					} else {
						if (remoteOutCounter >= Env.remoteReportInterval.getValue()) { 
							double curTime = System.currentTimeMillis();
							double timeTween = curTime - lastReportTime;
							lastReportTime = curTime;
							if (Env.logFiles) { 
								if (Env.makeFilLengthHisto.isActive()) { FilHistoControl.setFilLengthHisto(); }
								FileOps.writeToOutFile(); 
								FileOps.writeToFilLengthFile();
								FileOps.writeToFilLengthHistoFile();
								FileOps.writeToCapStateFile();
								FileOps.writeToFilLifetimeHistoFile();
								FileOps.writeToFilCapStateHistoFile();
							}
							talk ("Time: " + String.valueOf(Env.simulationTime));
							talk (" , timeTween: " + String.valueOf(timeTween));
							if (Env.cantileverElasticityTst) {
								talkln (" , Current Length is " + Env.tstFilLength + " , Current deflection is " + Env.currentDeflection + " , Expected deflection is " + Env.expectedDeflection);
							} else if (Env.testingPersistenceLength) {
								talkln ("");
							//	talkln (" , Length is " + ParMSpindle_Graphics.getPersistenceTstLengthString() + " , LastSegCorr is " + ParMSpindle_Graphics.getLastSegCorrelationString());
							} else {
								talk (" , ParM Conc: " + String.valueOf(Env.parMConc.getValue()));
								talk (" , ParM Conc (non-hydro): " + String.valueOf(Env.parMConcNonHydro.getValue()));
								//talk (" , MonsInSim: " + Bug.getMonsInSim());
								talk (" , # of Filament: " + String.valueOf(FilSegment.filCt));
								talk (" , # of Filament Segments: "  + String.valueOf(FilSegment.filSegmentCt));
								talk (" , # of Mons: " + String.valueOf(FilSegment.monomerSum()));
								talkln (" , total length: " + String.valueOf(FilSegment.lengthSum()));
							}
							remoteOutCounter = 0;
						}
						
						if (Env.cantileverElasticityTst & (ckElasticityCounter >= Env.ckElasticityInterval)) {
							elasticityCk();
							ckElasticityCounter = 0;
						}
						
						if (Env.simpleSupportElasticityTst & Env.simulationTime > 100) {
							FilSegment.simpleSupportRelaxActions();
						}
						
						if (Env.testingPersistenceLength & (ckPersistenceCounter >= Env.ckPersistenceInterval)) {
							FilSegment.persistenceTrack();
							FilSegment.transDiffTrack();
							ckPersistenceCounter = 0;
						}
						
						if ((Env.toQKFile) & (qkCounter >= Env.toQKFileInterval.getIntValue())) {
							FilSegment.updateAllMonomerPositions();
							FileOps.takeQuickPicture(Env.qkFilePath + File.separator + FileOps.makeNameFromIntStep(Env.outFileName,Env.counter) + ".qk");
							qkCounter = 0;
						}
					}
					
					// one rate annealing approx.
					if (Env.biochemOnly.isActive() & biochemOnlyActionsCounter >= Thing.biochemCheckInt) { 
						FilSegment.setAnnealRate();
						FilSegment.oneRateAnnealing(); 
						biochemOnlyActionsCounter = 0;
					}
					
					// remove Things AFTER the graphical update... can detach graphics objects of dead things this way
					Thing.removeDeadThings();		// get rid of dead things
					FilLink.removeDeadFilLinks();	// get rid of dissolved links
					
					// stochastically create new filaments
					if (FilSegment.filCt < Env.maxFils.getIntValue()) {
						if (Env.kRdmNuc.isActive()) { FilSegment.spawnRdmFilaments(); }
						if (Env.kPlasmidNuc.isActive()) { FilSegment.spawnPlasmidFilaments(); }
					}
				}
			}
		}
	}
	
	public static void makeInitialThings() {
		if (Env.fromQKStatePath != null) {
			FileOps.loadQuickState(Env.fromQKStatePath);
		} else {
			FilSegment.makeInitialFilaments();
			Plasmid.makeInitialPlasmids();
			//FilSegment.makeLinkTstFils();
			//FilSegment.plasmidsCrossedFilsTst();
			//FilSegment.plasmidsSingleFilTst();
			//FilSegment.crossedFilsTst();
			//FilSegment.annealTst();
			//FilSegment.cantileverElasticityTuning();
			//FilSegment.simpleSupportElasticityTuning();
			//FilSegment.persistenceLengthTst();
		}
	}
	
	public static void restartRun (boolean newParamsLoaded) {
		synchronized(Env.safeO) {
			setPaused();
			Env.setTimeStepCounts();	// reset time-step counters
			Monomer.removeAll();	// removes monomers from rendered files
			FilSegment.removeAll();
			Plasmid.removeAll();
			FilLink.removeAll();
			Thing.removeDeadThings();
			Thing.thingCt = 0;	// no Things left
			if (!Env.remote) { Thing.theBug.G.detach(); }
			Thing.theBug = null;
			Bug.makeInitialBug();
			if (Env.biochemOnly.isActive()) { FilSegment.resetOrderedY(); }
			Env.simulationTime = 0;
			
			if (Env.cantileverElasticityTst | Env.simpleSupportElasticityTst) { Env.resetElasticityTstValues(); }
			if (Env.testingPersistenceLength) { Env.resetPersistenceTstValues(); }
			
			if (Env.testLinkedPlasmids) {
				FilSegment.plasmidsSingleFilTst();
			} else {
				makeInitialThings();
			}
			
			if (!Env.remote) {
				if (newParamsLoaded) {
					if (!parMGraphics.canvasSizeCurrent()) { parMGraphics.makeCanvasSizeCurrent(); }
				} else {
					parMGraphics.updateCanvasSizeParams();
				}
				ParMSpindle_Graphics.updateBugScene();
				ParMSpindle_Graphics.updateAllControls();
				ParMSpindle_Graphics.theCanvas.resetFileCt();
			}
			if (Env.makeFilLengthHisto.isActive()) { FilHistoControl.setFilLengthHisto(); }
			
		}
	}
	
	public static void setRunning () {
		Env.paused = false;
		if (!Env.remote) {
			parMGraphics.pausedMenuItem.setState(false);
			parMGraphics.runMenuItem.setState(true);
		}
	}
	
	public static void setPaused () {
		Env.paused = true;
		if (!Env.remote) {
			parMGraphics.pausedMenuItem.setState(true);
			parMGraphics.runMenuItem.setState(false);
		}
	}
	
	public static void startUpdateTimer() {
		int delay = 1000; //milliseconds
		ActionListener taskPerformer = new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				parMGraphics.updateAllControls();
		    }
		};
		new Timer(delay, taskPerformer).start();
	}
	
	public static void elasticityCk () {
		Env.curDeltaD = Env.lastDeflection - Env.currentDeflection;
		Env.curDistFromExpected = Env.expectedDeflection - Env.currentDeflection;
		
		if (Math.abs(Env.curDistFromExpected) > Math.abs(Env.lastDistFromExpected)) {	// overshoot
			FileOps.talk ("Overshoot... ");
			changeElasticity();
		}
		if (Math.abs(Env.curDeltaD) < Env.ckElasticitySmallChange) { // if sufficiently static then check
			FileOps.talk ("Static deflection achieved...");
			if (Math.abs(Env.curDistFromExpected) < Env.valOKChange) {	// close enough to expected deflection
				FileOps.talkln (" done with this length");
				Env.elasReportString = "Final k value with filament length of " + deflectionFormat.format(Env.tstFilLength) + " is " + expFormat.format(Env.filTorqSpring.getValue());
				FileOps.talkln (" **** " + Env.elasReportString);
				FileOps.talkln ("");
				FilSegment curF;
				for (int i=0;i<FilSegment.filSegmentCt;i++) {
					curF = FilSegment.theFilSegments[i];
					if (!curF.filAtEnd2) {	// then this is the end filsegment to grow
						int monsToAdd = Env.stdSegLength.getIntValue();
						curF.monomerCt += monsToAdd;
						Monomer curMon = curF.plusMon;
						for (int j=1;j<monsToAdd;j++) {
							Monomer.polymerize(curMon,curF,Monomer.PLUSEND, true);
							curMon = curMon.frontMon;
						}
					}
				}
				Env.curMultipleOfSeg ++;
				Env.simulationTime = 0;
				Env.resetElasticityTstValues();
				if (Env.tstFilLength > Env.tstFilLengthMax) { 
					FileOps.talkln ("Done with elasticity test");
					System.exit(0); 
				}
			} else {
				changeElasticity();
			}
		}
		Env.lastDeflection = Env.currentDeflection;
		Env.lastDistFromExpected = Env.curDistFromExpected;
	}
	
	public static void changeElasticity () {
		if (Math.abs(Env.currentDeflection) > Math.abs(Env.expectedDeflection)) { 
			if (!Env.lastEChangePlus) {	// only halve elasticity change if we have switched direction of change
				Env.curDeltaE = Env.curDeltaE/2;
				Env.lastEChangePlus = true;
			}
			Env.filTorqSpring.setValue(Env.filTorqSpring.getValue()+Env.curDeltaE); 
			FileOps.talkln (" increasing spring to " + Env.filTorqSpring.getValue());
		} else {
			if (Env.lastEChangePlus) {
				Env.curDeltaE = Env.curDeltaE/2;
				Env.lastEChangePlus = false;
			}
			Env.filTorqSpring.setValue(Env.filTorqSpring.getValue()-Env.curDeltaE);
			FileOps.talkln (" decreasing spring to " +  + Env.filTorqSpring.getValue());
		}
	}
	
	/*public static void elasticityCk () {
		Env.curDeltaD = Env.lastDeflection - Env.currentDeflection;
		
		if (Math.abs(Env.curDeltaD) < Env.ckElasticitySmallChange) { // if sufficiently static then check
			talk ("Static deflection achieved...");
			if (Math.abs(Env.currentDeflection - Env.expectedDeflection) < Env.valOKChange) {	// close enough to expected deflection
				talkln (" done with this length");
				talkln (" **** Value with filament length of " + Env.tstFilLength + " is " + Env.filTorqSpring.getValue());
				talkln ("");
				FilSegment curF;
				for (int i=0;i<FilSegment.filSegmentCt;i++) {
					curF = FilSegment.theFilSegments[i];
					if (!curF.filAtEnd2) {	// then this is the end filsegment to grow
						int monsToAdd = Env.stdSegLength.getIntValue();
						curF.monomerCt += monsToAdd;
						Monomer curMon = curF.plusMon;
						for (int j=1;j<monsToAdd;j++) {
							Monomer.polymerize(curMon,curF,Monomer.PLUSEND, true);
							curMon = curMon.frontMon;
						}
					}
				}
				Env.curMultipleOfSeg ++;
				Env.simulationTime = 0;
				Env.resetElasticityTstValues();
				if (Env.tstFilLength > Env.tstFilLengthMax) { 
					talkln ("Done with elasticity test");
					System.exit(0); 
				}
			} else {
				if (Math.abs(Env.currentDeflection) > Math.abs(Env.expectedDeflection)) { 
					Env.filTorqSpring.setValue(Env.filTorqSpring.getValue()+1e-20); 
					talkln (" increasing spring.");
				} else {
					Env.filTorqSpring.setValue(Env.filTorqSpring.getValue()-1e-20);
					talkln (" decreasing spring.");
				}
			}
		}
		Env.lastDeflection = Env.currentDeflection;
		Env.lastDeltaD = Env.curDeltaD;
	} */
	
	private static void talkln (String info) {
		System.out.println(info);
	}
	
	private static void talk (String info) {
		System.out.print(info);
	}
}
