package parMSpindle;
/* ParMSpindle_Graphics.... Java3D and events, etc */

/*  <ParMSpindle - a simulation of ParM based plasmid segregation>
    Copyright (C) <2008>  <Jonathan B. Alberts>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import javax.media.j3d.*;
import javax.swing.JFrame;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.JSlider;
import javax.vecmath.*;
import javax.swing.event.*;

import java.io.File;
import java.text.DecimalFormat;

import com.sun.j3d.utils.universe.*;

public class ParMSpindle_Graphics extends JFrame
	implements WindowListener, CheckboxMenuListener, ActionListener, ItemListener, ChangeListener {

	//  Should the scene be compiled?
	private boolean shouldCompile = true;

	//  GUI objects for our subclasses
	protected Frame    theFrame    = null;
	protected MenuBar  theMenuBar  = null;
	protected static CapturingCanvas3D theCanvas = null;
	protected static JPanel topPanel = null;
	protected static JPanel bottomPanel = null;
	protected static JLabel timeLabel = null;
	protected static JLabel concLabel = null;
	protected static JLabel filCtLabel = null;
	protected static JLabel monCtLabel = null;
	protected TransformGroup theViewTransform = null;
	protected TransformGroup theSceneTransform = null;
	public JSlider bugTrans,bugEmit,headLightI,bugLightI;
	
	public static int frameHeadSize;		// different on different systems... size of bar at top of frame and the diff between canvas3d and JFrame height
	public static int bottomPanelHeight = 50;
	
	//  Private GUI objects and state
	public DirectionalLight headlight = null;
	public PointLight bugEnd1Light = null;
	public PointLight bugEnd2Light = null;
	private ExamineViewerBehavior examineBehavior = null;

	private Background background = null;
	private Switch shadingSwitch = null;
	private static Group scene = null;
	private static BranchGroup bugState = null;
	
	
	// for menus
	private static boolean headlightOn = true;
	private static boolean bugEndLightsOn = false;
	private static CheckboxMenuItem headlightMenuItem = null;
	private static CheckboxMenuItem bugEndLightsMenuItem = null;
	private static CheckboxMenuItem plasmidlightMenuItem = null;
	private CheckboxMenuItem paintOnMenuItem = null;
	private static CheckboxMenuItem info2DMenuItem = null;
	private static CheckboxMenuItem info3DMenuItem = null;
	private static CheckboxMenuItem vanityPlasMenuItem = null;
	private static CheckboxMenuItem writeFileMenuItem = null;
	private static CheckboxMenuItem writeQKFileMenuItem = null;
	private static CheckboxMenuItem renderQKFileMenuItem = null;
	public static CheckboxMenuItem pausedMenuItem = null;
	public static CheckboxMenuItem runMenuItem = null;
	private static CheckboxMenuItem rotateMenuItem = null;
	private static CheckboxMenuItem helixFilsMenuItem = null;
	private static CheckboxMenuItem sphereFilsMenuItem = null;
	private static CheckboxMenuItem filMotionOffMenuItem = null;
	private static CheckboxMenuItem plasmidMotionOffMenuItem = null;
	private static CheckboxMenuItem testLinkedPlasmidsMenuItem = null;
	private static CheckboxMenuItem compressionCritMenuItem = null;
	private static CheckboxMenuItem orderedCenteredMenuItem = null;
	private static CheckboxMenuItem filCoordSysOnMenuItem = null;
	private static CheckboxMenuItem bugOffMenuItem = null;
	private static CheckboxMenuItem relaxCantileverMenuItem = null;
	private static CheckboxMenuItem relaxSimpleSupportMenuItem = null;
	private static CheckboxMenuItem BugSlidersMenuItem = null;
	private int currentAppearance = 0;
	private CheckboxMenuItem[] appearanceMenu;
	
	// report strings
	static String filCtString;
	static String filSegCtString;
	
	
	// control frames with organizing tab and panel
	public static JFrame controlFrame = new JFrame("Control Panel");
	public static JPanel controlPanel = new JPanel(new GridLayout(1,1));
	public static JTabbedPane controlTab = new JTabbedPane();
	public static InitControl initControl;
	public static GraphicsControl graphicsControl;
	public static RenderControl renderControl;
	public static EnvControl envControl;
	public static MechControl mechControl;
	public static PlasmidControl plasControl;
	public static End1RatesControl end1RatesControl;
	public static End2RatesControl end2RatesControl;
	public static MiscRatesControl miscRatesControl;
	public static FilHistoControl filHistoControl;
	
	
	// text formats
	static DecimalFormat timeFormat = new DecimalFormat ("00.0000");
	static DecimalFormat concFormat = new DecimalFormat ("0.0000");
	static DecimalFormat deflectionFormat = new DecimalFormat ("0.0000");
	static DecimalFormat expFormat = new DecimalFormat ("0.000E0");
	
	public ParMSpindle_Graphics( ) {
		
		if (!Env.remote) { 
			makeAllControls(this); 
			new infoCCD.Info(true,true,true);
		}
		
	}
	
	public static void makeAllControls(ParMSpindle_Graphics pG) {
		initControl = new InitControl(pG);
		envControl = new EnvControl(pG);
		end1RatesControl = new End1RatesControl(pG);
		end2RatesControl = new End2RatesControl(pG);
		miscRatesControl = new MiscRatesControl(pG);
		mechControl = new MechControl(pG);
		plasControl = new PlasmidControl(pG);
		graphicsControl = new GraphicsControl(pG);
		filHistoControl = new FilHistoControl(pG);;
		
		controlTab.addTab("init",initControl.getContentPane());
		controlTab.addTab("env",envControl.getContentPane());
		controlTab.addTab("end1",end1RatesControl.getContentPane());
		controlTab.addTab("end2",end2RatesControl.getContentPane());
		controlTab.addTab("miscRates",miscRatesControl.getContentPane());
		controlTab.addTab("mech",mechControl.getContentPane());
		controlTab.addTab("plas",plasControl.getContentPane());
		controlTab.addTab("graphics",graphicsControl.getContentPane());
		controlTab.addTab("filHisto",filHistoControl.getContentPane());
		
		controlPanel.add(controlTab);
		controlFrame.getContentPane().add(controlPanel);
		// set location and size of controlFrame
		controlFrame.setSize(controlFrame.getPreferredSize());
		controlFrame.setLocationRelativeTo(null);
		//controlFrame.setLocation(Env.frameWidth.getIntValue()+10,0);
		
		renderControl = new RenderControl(pG);	// not part of param control set above
	}
	
	public static void updateAllControls() {
		initControl.updateInfoFields();
		graphicsControl.updateInfoFields();
		envControl.updateInfoFields();
		mechControl.updateInfoFields();
		end1RatesControl.updateInfoFields();
		end2RatesControl.updateInfoFields();
		miscRatesControl.updateInfoFields();
		filHistoControl.updateInfoFields();
	}


	/**
	 *  Initializes the Example by parsing command-line arguments,
	 *  building an AWT Frame, constructing a menubar, and creating
	 *  the 3D canvas.
	 *
	 *  @param   args    a String array of command-line arguments
	 */
	protected void initialize() {

		// Build the frame
		theFrame = new Frame( );
		theFrame.setBackground(Env.universeColor);
		theFrame.setSize( Env.frameWidth.getIntValue(), Env.frameHeight.getIntValue() );
		theFrame.setTitle( "Java 3D Example" );
		theFrame.setLayout( new BorderLayout( ) );

		// Set up a close behavior
		theFrame.addWindowListener( this );
		
		// Create a canvas
		theCanvas = new CapturingCanvas3D(SimpleUniverse.getPreferredConfiguration());
		theCanvas.setSize( Env.frameWidth.getIntValue(), Env.frameHeight.getIntValue() );
		theFrame.add( "Center", theCanvas );

		// Create time panel
		//makeTopPanel();
		//theFrame.add( "North", topPanel );
		
		// Create bottom panel
		makeBottomPanel();
		//theFrame.add("South",bottomPanel);
		
		// Pack
		theFrame.pack( );
		theFrame.validate( );
		//		exampleFrame.setVisible( true );
		
		theFrame.setTitle( "ParM Spindle Simulation" );
		makeMenuBar();
		
		// set frameHeadSize for future canvas/jframe resizing
		frameHeadSize = theFrame.getHeight() - theCanvas.getHeight();
		//talkln ("frameHeadSize: " + frameHeadSize);
		
	}



	//--------------------------------------------------------------
	//  SCENE CONTENT
	//--------------------------------------------------------------

	/**
	 *  Builds the 3D universe by constructing a virtual universe
	 *  (via SimpleUniverse), a view platform (via SimpleUniverse), and
	 *  a view (via SimpleUniverse).  A headlight is added and a
	 *  set of behaviors initialized to handle navigation types.
	 */
	protected void buildUniverse( ) {
		SimpleUniverse universe = new SimpleUniverse(theCanvas);
		
		Viewer viewer = universe.getViewer( );
		ViewingPlatform viewingPlatform = universe.getViewingPlatform( );
		theViewTransform = viewingPlatform.getViewPlatformTransform( );


		//
		//  Create a "headlight" as a forward-facing directional light.
		//  Set the light's bounds to huge.  Since we want the light
		//  on the viewer's "head", we need the light within the
		//  TransformGroup containing the ViewPlatform.  The
		//  ViewingPlatform class creates a handy hook to do this
		//  called "platform geometry".  The PlatformGeometry class is
		//  subclassed off of BranchGroup, and is intended to contain
		//  a description of the 3D platform itself... PLUS a headlight!
		//  So, to add the headlight, create a new PlatformGeometry group,
		//  add the light to it, then add that platform geometry to the
		//  ViewingPlatform.
		//
		BoundingSphere allBounds = new BoundingSphere(
			new Point3d( 0.0, 0.0, 0.0 ), 100000.0 );

		PlatformGeometry pg = new PlatformGeometry( );
		headlight = new DirectionalLight( );
		headlight.setColor( Env.headlightColor );
		headlight.setDirection( new Vector3f( 0.0f, 0.0f, -1.0f ) );
		headlight.setInfluencingBounds( allBounds );
		headlight.setCapability( Light.ALLOW_STATE_WRITE );
		headlight.setCapability( Light.ALLOW_COLOR_WRITE );
		pg.addChild( headlight );
		
		bugEnd1Light = new PointLight();
		bugEnd1Light.setCapability(Light.ALLOW_STATE_WRITE);
		bugEnd1Light.setCapability(Light.ALLOW_COLOR_WRITE);
		bugEnd1Light.setColor(Env.bugEndLightColor3f);
		bugEnd1Light.setInfluencingBounds(allBounds);
		bugEnd1Light.setAttenuation(5.0f,0f,0.0f);
		bugEnd1Light.setPosition(-1f,0f,0f);
		bugEnd1Light.setEnable(bugEndLightsOn);
		pg.addChild(bugEnd1Light);
		
		bugEnd2Light = new PointLight();
		bugEnd2Light.setCapability(Light.ALLOW_STATE_WRITE);
		bugEnd2Light.setCapability(Light.ALLOW_COLOR_WRITE);
		bugEnd2Light.setColor(Env.bugEndLightColor3f);
		bugEnd2Light.setInfluencingBounds(allBounds);
		bugEnd2Light.setAttenuation(5.0f,0f,0.0f);
		bugEnd2Light.setPosition(1f,0f,0f);
		bugEnd2Light.setEnable(bugEndLightsOn);
		pg.addChild(bugEnd2Light);
		
		viewingPlatform.setPlatformGeometry( pg );

		//
		//  Create the 3D content BranchGroup, containing:
		//
		//    - a TransformGroup who's transform the examine behavior
		//      will change (when enabled).
		//
		//    - 3D geometry to view
		//
		// Build the scene root
		BranchGroup sceneRoot = new BranchGroup( );

		// Build a transform that we can modify
		theSceneTransform = new TransformGroup( );
		theSceneTransform.setCapability(TransformGroup.ALLOW_TRANSFORM_READ );
		theSceneTransform.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE );
		theSceneTransform.setCapability(Group.ALLOW_CHILDREN_EXTEND );


		//
		//  Build the scene, add it to the transform, and add
		//  the transform to the scene root
		//
		Group scene = this.buildScene( );
		theSceneTransform.addChild( scene );
		sceneRoot.addChild( theSceneTransform );


		examineBehavior = new ExamineViewerBehavior(theSceneTransform, theFrame ); 
		examineBehavior.setSchedulingBounds( allBounds );
		sceneRoot.addChild( examineBehavior );

		examineBehavior.setEnable( true );


		//
		//  Compile the scene branch group and add it to the SimpleUniverse.
		//
		if ( shouldCompile ) { sceneRoot.compile( ); }
		universe.addBranchGraph( sceneRoot );
		resetView();
		
		// antialias
		theCanvas.getView().setSceneAntialiasingEnable(true);
	}
	
	public boolean canvasSizeCurrent () {
		if (theCanvas.getHeight() != Env.frameHeight.getIntValue()) { return false; }
		if (theCanvas.getWidth() != Env.frameWidth.getIntValue()) { return false; }
		return true;
	}
	
	public void makeCanvasSizeCurrent () {
		setCanvasSize (Env.frameWidth.getIntValue(), Env.frameHeight.getIntValue());
	}
	
	public void updateCanvasSizeParams() {
		if (BugSlidersMenuItem.getState( )) {
			Env.frameWidth.setValue(theFrame.getWidth());
			Env.frameHeight.setValue(theFrame.getHeight()-frameHeadSize);
		} else {
			Env.frameWidth.setValue(theFrame.getWidth());
			Env.frameHeight.setValue(theFrame.getHeight()-frameHeadSize-bottomPanelHeight);
		}
	}

	public void setCanvasSize (int width, int height) {
		//talkln ("Resetting canvas size to " + width + "," + height);
		theCanvas.stopRenderer();
		theFrame.remove(theCanvas);
		theFrame.setSize(width,height+frameHeadSize);
		theCanvas.setSize(width,height);
		theFrame.add("Center",theCanvas);
		theCanvas.startRenderer();
		
		//topPanel.setSize(width,height);
		//topPanel.remove(timeLabel);
		//topPanel.add(timeLabel);
		//topPanel.add(concLabel);
	}
	
	//  Reset transforms
	public void resetView( )
	{
		Transform3D trans = new Transform3D( );
		trans.setScale(Env.transScale);
		theSceneTransform.setTransform( trans );
		trans.set( new Vector3f( 0.0f, 0.0f, 30.0f ) );
		theViewTransform.setTransform( trans );
	}
	
	public void setRotZ90View( )
	{
		Transform3D trans = new Transform3D( );
		Transform3D rotTrans = new Transform3D();
		rotTrans.rotZ(Math.PI/2);
		trans.mul(rotTrans);
		trans.setScale(Env.transScale);
		theSceneTransform.setTransform( trans );
		trans.set( new Vector3f( 0.0f, 0.0f, 30.0f ) );
		theViewTransform.setTransform( trans );
	}
	
	public void rotateView() {
		double rotAng = 1*Math.PI/180;
		Transform3D rotT3D = new Transform3D();
		rotT3D.rotY(rotAng);
		Transform3D curT3D = new Transform3D();
		theSceneTransform.getTransform(curT3D);
		curT3D.mul(rotT3D);
		curT3D.setScale(Env.transScale);
		theSceneTransform.setTransform(curT3D);
	}
	
	public void rotateView(double rotAng) {
		rotAng = rotAng*Math.PI/180;
		Transform3D rotT3D = new Transform3D();
		rotT3D.rotY(rotAng);
		Transform3D curT3D = new Transform3D();
		theSceneTransform.getTransform(curT3D);
		curT3D.mul(rotT3D);
		curT3D.setScale(Env.transScale);
		theSceneTransform.setTransform(curT3D);
	}


	private BranchGroup initialBugScene() {
		bugState = new BranchGroup();
		allBranchGroupCapabilities(bugState);
		
		for (int i=0;i<Thing.thingCt;i++) {
			bugState.addChild(Thing.theThings[i].getGraphicsNode());
		}
		return bugState;
	}
	
	public static void updateBugScene() {
		Thing curThing;
		FilLink curFL;
		//if (bugState != null) { bugState.detach();}
		//bugState = new BranchGroup();
		//allBranchGroupCapabilities(bugState);

		for (int i=0;i<Thing.thingCt;i++) {
			curThing = Thing.theThings[i];
			if (!curThing.graphicsMade) {
				bugState.addChild(curThing.getGraphicsNode());
			} else {
				curThing.updateGraphics();
			}
		}
		for (int i=0;i<FilLink.filLinkCt;i++) {
			curFL = FilLink.filLinks[i];
			if (!curFL.graphicsMade) {
				bugState.addChild(curFL.getGraphicsNode());
			} else {
				curFL.updateGraphics();
			}
		}
		//scene.addChild(bugState);
	}
	
	public static void updateQKBugScene() {
		Thing curThing;
		FilLink curFL;
		Monomer curMon;
		
		if (Env.filRenderOff) {
			for (int i=0;i<Plasmid.plasmidCt+1;i++) {
				curThing = Thing.theThings[i];
				if (!curThing.graphicsMade) {
					bugState.addChild(curThing.getGraphicsNode());
				} else {
					curThing.updateGraphics();
				}
			}
			return;
		} 
		

		for (int i=0;i<Thing.thingCt;i++) {
			curThing = Thing.theThings[i];
			if (!curThing.graphicsMade) {
				bugState.addChild(curThing.getGraphicsNode());
			} else {
				curThing.updateGraphics();
			}
		}
		
		for (int i=0;i<FilLink.filLinkCt;i++) {
			curFL = FilLink.filLinks[i];
			if (!curFL.graphicsMade) {
				bugState.addChild(curFL.getGraphicsNode());
			} else {
				curFL.updateGraphics();
			}
		}
		
		for (int x=Monomer.monRenderCt;x<Monomer.monCt;x++) {
			Monomer.theMonomers[x].detachGraphics();
		}
		
		for (int m=0;m<Monomer.monRenderCt;m++) {
			curMon = Monomer.theMonomers[m];
			//talkln ("curMon number " + String.valueOf(m) + " is " + String.valueOf(curMon) + " and is in state " + String.valueOf(curMon.getState()));
			if (!curMon.graphicsIn) {
				curMon.addGraphics(bugState);
			}
		}
	}
	
	
	public static void clearForRun () {
		synchronized(Env.safeO) {
			Monomer.removeAll();	// removes monomers from rendered files
			FilSegment.removeAll();
			Plasmid.removeAll();
			Thing.removeDeadThings();
			Thing.thingCt = 0;	// no Things left
			if (!Env.remote) { Thing.theBug.G.detach(); }
			Thing.theBug = null;
			if (!Env.remote) {
				//updateBugScene();
				updateAllControls();
			}
			Env.simulationTime = 0;
		}
	}
	
	public static void clearForRender () {
		synchronized(Env.safeO) {
			FilSegment.removeAll();
			Plasmid.removeAll();
			Thing.removeDeadThings();
			Thing.thingCt = 0;	// no Things left
			Thing.theBug.G.detach();
			Thing.theBug = null;
		}
	}
	
	public static void captureImage (CapturingCanvas3D theCanvas) {
		theCanvas.writeJPEG_ = true;
		theCanvas.repaint();
	}
	
	public static void waitOnCapture (CapturingCanvas3D theCanvas) {
		while (theCanvas.isRendering()) {
			try { Thread.sleep(500); } catch (InterruptedException e) { talkln ("error sleeping"); }
		}
	}
	
	//
	//  Build scene
	//
	public Group buildScene( )
	{
		
		// Build the scene root
		scene = new Group( );
		scene.setCapability(scene.ALLOW_CHILDREN_WRITE);
		scene.setCapability(scene.ALLOW_CHILDREN_EXTEND);

		// Create application bounds
		BoundingSphere worldBounds = new BoundingSphere(
			new Point3d( 0.0, 0.0, 0.0 ),  // Center
			1000.0 );                      // Extent

		// Set the background color and its application bounds
		background = new Background( );
		background.setColor(Env.universeColor3f);
		background.setCapability( Background.ALLOW_IMAGE_WRITE );
		background.setApplicationBounds( worldBounds );
		scene.addChild( background );

		// Build foreground geometry
		scene.addChild( initialBugScene( ) );

		return scene;
	}


	public void makeTopPanel () {
		topPanel = new JPanel();
		topPanel.setSize(Env.frameWidth.getIntValue(),8);
		topPanel.setLayout(new GridLayout(1,4));
		topPanel.setBackground(Env.controlBackColor);
		
		timeLabel = new JLabel(" ");
		timeLabel.setForeground(Env.controlForeColor);
		topPanel.add(timeLabel);
		
		concLabel = new JLabel(" ");
		concLabel.setForeground(Env.controlForeColor);
		topPanel.add(concLabel);
		
		filCtLabel = new JLabel(" ");
		filCtLabel.setForeground(Env.controlForeColor);
		topPanel.add(filCtLabel);
		
		monCtLabel = new JLabel(" ");
		monCtLabel.setForeground(Env.controlForeColor);
		topPanel.add(monCtLabel);
		
	}
	
	public void makeBottomPanel() {
		bottomPanel = new JPanel();
		bottomPanel.setSize(Env.frameWidth.getIntValue(),bottomPanelHeight);
		bottomPanel.setLayout(new GridLayout(2,2));
		//bottomPanel.setBackground(Env.controlBackColor);
		JPanel subP1 = new JPanel();
		JPanel subP2 = new JPanel();
		JPanel subP3 = new JPanel();
		JPanel subP4 = new JPanel();
		
		JLabel bugTransLabel = new JLabel("   Bug Transparency:");
		bugTransLabel.setForeground(Env.canvasTextColor);
		bugTrans =  new JSlider(JSlider.HORIZONTAL,0,100,(int)(100*Bug.initialTrans));
		bugTrans.addChangeListener(this);
		subP1.add(bugTransLabel);
		subP1.add(bugTrans);
		
		JLabel bugEmitLabel = new JLabel(" Bug Emissive Color:");
		bugEmitLabel.setForeground(Env.canvasTextColor);
		bugEmit = new JSlider(JSlider.HORIZONTAL,0,100,(int)(100*Bug.initialEmit));
		bugEmit.addChangeListener(this);
		subP2.add(bugEmitLabel);
		subP2.add(bugEmit);
		
		JLabel headLightILabel = new JLabel("HeadLight Intensity:");
		headLightILabel.setForeground(Env.canvasTextColor);
		headLightI = new JSlider(JSlider.HORIZONTAL,0,100,(int)(100*Env.headlightColor.x));
		headLightI.addChangeListener(this);
		subP3.add(headLightILabel);
		subP3.add(headLightI);
		
		JLabel bugLightILabel = new JLabel("   BugLight Intensity:");
		bugLightILabel.setForeground(Env.canvasTextColor);
		bugLightI = new JSlider(JSlider.HORIZONTAL,0,100,(int)(100*Env.bugEndLightColor3f.x));
		bugLightI.addChangeListener(this);
		subP4.add(bugLightILabel);
		subP4.add(bugLightI);
		
		bottomPanel.add(subP1);
		bottomPanel.add(subP2);
		bottomPanel.add(subP3);
		bottomPanel.add(subP4);
	}
	
	public static void setTopLabels () {
		setTimeLabel(); 
		setConcLabel();
		setFilCtLabel();
		setMonCtLabel();
	}
	
	public static String getTimeString() {
		return new String (" Time: " + String.valueOf(timeFormat.format(Env.simulationTime)) + " s ");
	}
	
	public static void setTimeLabel () {
		timeLabel.setText(getTimeString());
	}
	
	public static String getConcString() {
		return new String (" ParM: " + String.valueOf(concFormat.format(Env.parMConc.getValue())) + " M "); 
	}
	
	public static String getNonHydroConcString() {
		return new String (" Non-hydro ParM: " + String.valueOf(concFormat.format(Env.parMConcNonHydro.getValue())) + " M "); 
	}
	
	public static void setConcLabel () {
		concLabel.setText(getConcString()); 
	}
	
	public static String getFilLinkCtString() {
		return new String (" FilLinks#: " + String.valueOf(FilLink.filLinkCt)); 
	}
	
	public static String getFilSegCtString() {
		int filCt;
		if (Env.fromQKFile) {
			filCt = FilSegment.filamentRenderCt;
		} else {
			filCt = FilSegment.filSegmentCt;
		}
		return new String (" FilSegs#: " + String.valueOf(filCt)); 
	}
	
	public static String getFilCtString() {
		int filCt;
		if (Env.fromQKFile) {
			filCt = FilSegment.filamentRenderCt;
		} else {
			filCt = FilSegment.filCt;
		}
		filCtString = " Filament#: " + String.valueOf(filCt);
		return filCtString; 
	}
	
	public static void setFilCtLabel () {
		filCtLabel.setText(getFilCtString());
	}
	
	public static String getMonCtString() {
		int monCt;
		if (Env.fromQKFile) {
			monCt = Monomer.monRenderCt;
		} else {
			monCt = FilSegment.monomerSum();
		}
		return new String (" Mon#: " + String.valueOf(monCt)); 
	}
	
	public static void setMonCtLabel () {
		monCtLabel.setText(getMonCtString());
	}
	
	public static String getTstFilLengthString() {
		return new String (" Length: " + String.valueOf(deflectionFormat.format(Env.tstFilLength)) + " M "); 
	}
	
	public static String getCurrentDeflectionString() {
		return new String ("Current : " + String.valueOf(deflectionFormat.format(Env.currentDeflection)) + " M "); 
	}
	
	public static String getExpectedDeflectionString() {
		return new String ("Expected : " + String.valueOf(deflectionFormat.format(Env.expectedDeflection)) + " M "); 
	}
	
	public static String getRelaxationDispString() {
		if (!Env.relaxCantilever & !Env.relaxSimpleSupport) { Env.relaxationDeflection = Env.currentDeflection/Math.E; }
		return new String ("/e: " + String.valueOf(deflectionFormat.format(Env.relaxationDeflection)) + " M "); 
	}
	
	public static String getCurrentElasticity() {
		return new String ("kTorsion: " + String.valueOf(expFormat.format(Env.filTorqSpring.getValue())) + " N/rad ");
	}
	
	public static String getPersistenceTstLengthString () {
		return new String ("Length: " + String.valueOf(deflectionFormat.format(Env.persistenceTstLength)) + " M "); 
	}
	
	public static String getLastSegCorrelationString () {
		double lastSegCorr = 0;
		if (Env.counter > 0) { 
			double scaleVal = 1/Env.persistenceTrack[0].runningAverageVal();		// in case we haven't filled valuetracker yet
			lastSegCorr = scaleVal*Env.persistenceTrack[Env.persistenceSegNumber-1].runningAverageVal(); 
		}
		return new String (" LastSegCorrelation: " + String.valueOf(deflectionFormat.format(lastSegCorr))); 
	}
	
	public void makeMenuBar () {
		theMenuBar = new MenuBar();
		theFrame.setMenuBar( theMenuBar );

		// Add menu for program control
		Menu m = new Menu ("Program");
		m.addActionListener(this);
		
		runMenuItem = new CheckboxMenuItem ("Run",!Env.paused);
		runMenuItem.addItemListener(this);
		m.add(runMenuItem);
		
		pausedMenuItem = new CheckboxMenuItem ("Pause",Env.paused);
		pausedMenuItem.addItemListener(this);
		m.add(pausedMenuItem);
		
		m.add("Restart");
		
		m.add("--------");
		
		writeFileMenuItem = new CheckboxMenuItem("Write JPEGs");
		writeFileMenuItem.addItemListener(this);
		writeFileMenuItem.setState(Env.toFile);
		m.add(writeFileMenuItem);
		
		writeQKFileMenuItem = new CheckboxMenuItem("Write QK files");
		writeQKFileMenuItem.addItemListener(this);
		writeQKFileMenuItem.setState(Env.toQKFile);
		m.add(writeQKFileMenuItem);
		
		m.add("--------");
		
		m.add("Load State from QK File");
		
		m.add("--------");
		
		m.add("Save Parameter Settings");
		m.add("Load Parameter Settings");
		
		theMenuBar.add(m);

		// View menu
		m = new Menu( "View" );
		m.addActionListener( this );

		m.add( "Reset view" );
		
		m.add( "Reset view - RotZ90");

		m.addSeparator( );

		headlightMenuItem = new CheckboxMenuItem( "Headlight On" );
		headlightMenuItem.addItemListener( this );
		headlightMenuItem.setState( headlightOn );
		m.add( headlightMenuItem );
		
		bugEndLightsMenuItem = new CheckboxMenuItem( "Bug End Lights On" );
		bugEndLightsMenuItem.addItemListener( this );
		bugEndLightsMenuItem.setState( bugEndLightsOn );
		m.add( bugEndLightsMenuItem );

		plasmidlightMenuItem = new CheckboxMenuItem( "Plasmid Lights On" );
		plasmidlightMenuItem.addItemListener( this );
		plasmidlightMenuItem.setState( true );
		m.add( plasmidlightMenuItem );
		
		theMenuBar.add( m );
		
		// Add a menu to select among options for painting etc
		m = new Menu( "Paint_Options" );
		m.add("---General---");
		
		paintOnMenuItem = new CheckboxMenuItem("Paint On");
		paintOnMenuItem.addItemListener(this);
		paintOnMenuItem.setState(Env.paintOn);
		m.add(paintOnMenuItem);
		
		rotateMenuItem = new CheckboxMenuItem("Rotate");
		rotateMenuItem.addItemListener(this);
		rotateMenuItem.setState(Env.viewRotation);
		m.add(rotateMenuItem);
		
		m.add("---Text Info---");
		
		info2DMenuItem = new CheckboxMenuItem("2D");
		info2DMenuItem.addItemListener(this);
		info2DMenuItem.setState(Env.infoIn2D);
		m.add(info2DMenuItem);
		
		info3DMenuItem = new CheckboxMenuItem("3D");
		info3DMenuItem.addItemListener(this);
		info3DMenuItem.setState(Env.infoIn3D);
		m.add(info3DMenuItem);
		
		m.add("---Bug---");

		appearanceMenu = new CheckboxMenuItem[3];
		appearanceMenu[0] = new CheckboxMenuItem( "Shaded" );
		appearanceMenu[0].addItemListener( this );
		if (Env.bugWiredCoarse) { 
			appearanceMenu[0].setState( false );
		} else {
			appearanceMenu[0].setState( !Env.bugWired );
		}
		m.add( appearanceMenu[0] );

		appearanceMenu[1] = new CheckboxMenuItem( "Wireframe" );
		appearanceMenu[1].addItemListener( this );
		if (Env.bugWiredCoarse) { 
			appearanceMenu[1].setState( false );
		} else {
			appearanceMenu[1].setState( Env.bugWired );
		}
		m.add( appearanceMenu[1] );
		
		appearanceMenu[2] = new CheckboxMenuItem( "Coarse Wireframe" );
		appearanceMenu[2].addItemListener( this );
		appearanceMenu[2].setState( Env.bugWiredCoarse );
		m.add( appearanceMenu[2] );
		
		m.add("---Plasmids---");
		
		vanityPlasMenuItem = new CheckboxMenuItem("Vanity Plasmid");
		vanityPlasMenuItem.addItemListener(this);
		vanityPlasMenuItem.setState(Env.plasmidFaces);
		m.add(vanityPlasMenuItem);
		
		m.add("---Filaments---");
		
		helixFilsMenuItem = new CheckboxMenuItem("Wire Helix");
		helixFilsMenuItem.addItemListener(this);
		helixFilsMenuItem.setState(!Env.helixSpheres);
		m.add(helixFilsMenuItem);
		
		sphereFilsMenuItem = new CheckboxMenuItem("Spherical Monomers");
		sphereFilsMenuItem.addItemListener(this);
		sphereFilsMenuItem.setState(Env.helixSpheres);
		m.add(sphereFilsMenuItem);
		
		theMenuBar.add( m );
		
		m = new Menu ( "Panels" );
		m.addActionListener( this );
		
		m.add("Control Panel");
		m.add("Render from QK files");
		
		BugSlidersMenuItem = new CheckboxMenuItem("Bug and Light Sliders");
		BugSlidersMenuItem.addItemListener(this);
		BugSlidersMenuItem.setState(false);
		m.add(BugSlidersMenuItem);
		
		theMenuBar.add( m );
		
		m = new Menu ( "Testing" );
		
		filMotionOffMenuItem = new CheckboxMenuItem("Filament Brownian Motion Off");
		filMotionOffMenuItem.addItemListener(this);
		filMotionOffMenuItem.setState(Env.brownianFilMotionOff);
		m.add(filMotionOffMenuItem);
		
		plasmidMotionOffMenuItem = new CheckboxMenuItem("Plasmid Brownian Motion Off");
		plasmidMotionOffMenuItem.addItemListener(this);
		plasmidMotionOffMenuItem.setState(Env.brownianFilMotionOff);
		m.add(plasmidMotionOffMenuItem);
		
		testLinkedPlasmidsMenuItem = new CheckboxMenuItem("Test Linked Plasmids");
		testLinkedPlasmidsMenuItem.addItemListener(this);
		testLinkedPlasmidsMenuItem.setState(Env.testLinkedPlasmids);
		m.add(testLinkedPlasmidsMenuItem);
		
		compressionCritMenuItem = new CheckboxMenuItem("Compression doesn't affect polymerization");
		compressionCritMenuItem.addItemListener(this);
		compressionCritMenuItem.setState(Env.compressionCritOff);
		m.add(compressionCritMenuItem);
		
		orderedCenteredMenuItem = new CheckboxMenuItem("Ordered Filaments Are Centered");
		orderedCenteredMenuItem.addItemListener(this);
		orderedCenteredMenuItem.setState(Env.orderedCentered);
		m.add(orderedCenteredMenuItem);
		
		filCoordSysOnMenuItem = new CheckboxMenuItem("Render Coord. Sys. on Segments");
		filCoordSysOnMenuItem.addItemListener(this);
		filCoordSysOnMenuItem.setState(Env.filCoordSysOn);
		m.add(filCoordSysOnMenuItem);
		
		bugOffMenuItem = new CheckboxMenuItem("Remove Bug from Simulation");
		bugOffMenuItem.addItemListener(this);
		bugOffMenuItem.setState(Env.bugOff);
		m.add(bugOffMenuItem);
		
		relaxCantileverMenuItem = new CheckboxMenuItem("Relax Cantilever");
		relaxCantileverMenuItem.addItemListener(this);
		relaxCantileverMenuItem.setState(Env.relaxCantilever);
		m.add(relaxCantileverMenuItem);
		
		relaxSimpleSupportMenuItem = new CheckboxMenuItem("Relax SimpleSupport");
		relaxSimpleSupportMenuItem.addItemListener(this);
		relaxSimpleSupportMenuItem.setState(Env.relaxSimpleSupport);
		m.add(relaxSimpleSupportMenuItem);
		
		theMenuBar.add(m);
		
		theMenuBar.add(infoCCD.Info.infoMenu);
		
	}
	
	public void actionPerformed( ActionEvent event ) {
		String arg = event.getActionCommand( );
		
		if ( arg.equals( "Reset view" ) ) {
			resetView();
		}
		
		if ( arg.equals( "Reset view - RotZ90") ) {
			setRotZ90View();
		}
	
		if (arg.equals("Restart")) { 
			ParMSpindle.setPaused();
			ParMSpindle.restartRun(false); 
		}
		
		if ( arg.equals( "Control Panel" ) ) {
			controlFrame.setVisible(true);
		}
		
		if ( arg.equals( "Render from QK files" ) ) {
			renderControl.setVisible(true);
		}
		
		if ( arg.equals("Load State from QK File") ) {
			prepForQKStateLoad();
			if (Env.fromQKStatePath != null) {
				FileOps.loadQuickState(Env.fromQKStatePath);
			}
		}
		
		if ( arg.equals("Save Parameter Settings") ) {
			saveParamsToFile();
		}
		
		if ( arg.equals("Load Parameter Settings") ) {
			loadParamsFromFile();
		}
		
	}


	//
	//  Handle checkboxes
	//
	public void itemStateChanged( ItemEvent event )
	{
		Object src = event.getSource( );

		boolean state;
		if ( src == headlightMenuItem )
		{
			state = headlightMenuItem.getState( );
			headlight.setEnable( state );
		}
		
		if ( src == bugEndLightsMenuItem ) {
			state = bugEndLightsMenuItem.getState( );
			bugEnd1Light.setEnable( state );
			bugEnd2Light.setEnable(state);
		}
		
		if ( src == plasmidlightMenuItem ) {
			state = plasmidlightMenuItem.getState( );
			Plasmid.plasmidsLightSwitch(state);
		}
		
		if ( src == BugSlidersMenuItem ) {
			updateCanvasSizeParams();
			boolean showSliders = BugSlidersMenuItem.getState( );
			if (showSliders) {
				theFrame.add("South",bottomPanel);
				theFrame.setSize(Env.frameWidth.getIntValue(),Env.frameHeight.getIntValue() + frameHeadSize + bottomPanelHeight);
			} else {
				theFrame.remove(bottomPanel);
				theFrame.setSize(Env.frameWidth.getIntValue(),Env.frameHeight.getIntValue() + frameHeadSize);
			}
		}
		
		// Check if it is an appearance choice
		if ( src == appearanceMenu[0] ) {
			appearanceMenu[0].setState( true );
			appearanceMenu[1].setState( false );
			appearanceMenu[2].setState( false );
			Env.bugWired = false;
			Env.bugWiredCoarse = false;
			Thing.theBug.appearanceChange();
			return;
		}
		if ( src == appearanceMenu[1] ) {
			appearanceMenu[0].setState( false );
			appearanceMenu[1].setState( true );
			appearanceMenu[2].setState( false );
			Env.bugWired = true;
			Env.bugWiredCoarse = false;
			Thing.theBug.appearanceChange();
			return;
		}
		
		if ( src == appearanceMenu[2] ) {
			appearanceMenu[0].setState( false );
			appearanceMenu[1].setState( false );
			appearanceMenu[2].setState( true );
			Env.bugWiredCoarse = true;
			Thing.theBug.appearanceChange();
			return;
		}
		
		if ( src == paintOnMenuItem) {
			Env.paintOn = paintOnMenuItem.getState();
		}
		
		if ( src == info2DMenuItem) {
			Env.infoIn2D = info2DMenuItem.getState();
		}
		
		if ( src == info3DMenuItem) {
			theCanvas.stopRenderer();
			Env.infoIn3D = info3DMenuItem.getState();
			Thing.theBug.updateTextInfoState();
			try { Thread.sleep(20); } catch (InterruptedException e) { talkln ("error sleeping"); }
			theCanvas.startRenderer();
		}
		
		if (src == writeFileMenuItem) {
			if (writeFileMenuItem.getState()) { 
				prepForJPEGWriting(); 
			} else {
				Env.toFile = false;
				Env.toFilePath = null;
				Env.toFileName = null;
			}
		}
		
		if (src == writeQKFileMenuItem) {
			if (writeQKFileMenuItem.getState()) { 
				prepForQKWriting(); 
			} else {
				Env.toQKFile = false;
				Env.qkFilePath = null;
				Env.outFileName = null;
			}
		}
		
		if (src == pausedMenuItem) {
			ParMSpindle.setPaused();	
		}
		
		if (src == runMenuItem) {
			ParMSpindle.setRunning();
		}
		
		if (src == rotateMenuItem) {
			Env.viewRotation = rotateMenuItem.getState();
		}
		
		if (src == sphereFilsMenuItem) {
			if (sphereFilsMenuItem.getState() != Env.helixSpheres) {
				Env.helixSpheres = sphereFilsMenuItem.getState();
				helixFilsMenuItem.setState(!Env.helixSpheres);
				FilSegment.resetGraphicsAll();
			}
		}
		
		if (src == helixFilsMenuItem) {
			if (Env.helixSpheres != !helixFilsMenuItem.getState()) {
				Env.helixSpheres = !helixFilsMenuItem.getState();
				sphereFilsMenuItem.setState(Env.helixSpheres);
				FilSegment.resetGraphicsAll();
			}
		}
		
		if (src == vanityPlasMenuItem) {
			Env.plasmidFaces = vanityPlasMenuItem.getState();
			Plasmid.switchLightCenters();
		}
		
		if (src == filMotionOffMenuItem) {
			Env.brownianFilMotionOff = filMotionOffMenuItem.getState();
		}
		
		if (src == plasmidMotionOffMenuItem) {
			Env.plasmidBrownianMotionOff = plasmidMotionOffMenuItem.getState();
		}
		
		if (src == compressionCritMenuItem) {
			Env.compressionCritOff = compressionCritMenuItem.getState();
		}
		
		if (src == testLinkedPlasmidsMenuItem) {
			Env.testLinkedPlasmids = testLinkedPlasmidsMenuItem.getState();
			ParMSpindle.restartRun(false);
		}
		
		if (src == orderedCenteredMenuItem) {
			Env.orderedCentered = orderedCenteredMenuItem.getState();
		}
		
		if (src == filCoordSysOnMenuItem) {
			Env.filCoordSysOn = filCoordSysOnMenuItem.getState();
		}
		
		if (src == bugOffMenuItem) {
			Env.bugOff = bugOffMenuItem.getState();
		}
		
		if (src == relaxCantileverMenuItem) {
			Env.relaxCantilever = relaxCantileverMenuItem.getState();
		}
		
		if (src == relaxSimpleSupportMenuItem) {
			Env.relaxSimpleSupport = relaxSimpleSupportMenuItem.getState();
			FilSegment.simpleSupportRelaxActions();
		}
	}
	
	public void stateChanged(ChangeEvent e) {
	    JSlider source = (JSlider)e.getSource();
	    if (source == bugTrans) {
		    if (!source.getValueIsAdjusting()) {
		       float newBugTrans = (float)(source.getValue())/100;
		       Bug.tA.setTransparency(newBugTrans);
		       Bug.appearanceChange();
		    }
	    }
	    
	    if (source == bugEmit) {
	    	if (!source.getValueIsAdjusting()) {
		       float newBugEmit = (float)(source.getValue())/100;
		       Thing.theBug.m.setEmissiveColor(newBugEmit,newBugEmit,newBugEmit);
		       Bug.appearanceChange();
		    }
	    }
	    
	    if (source == headLightI) {
	    	if (!source.getValueIsAdjusting()) {
		       float newHeadLightI = (float)(source.getValue())/100;
		       headlight.setColor(new Color3f(newHeadLightI,newHeadLightI,newHeadLightI));
		    }
	    }
	    
	    if (source == bugLightI) {
	    	if (!source.getValueIsAdjusting()) {
		       float newBugLightI = (float)(source.getValue())/100;
		       bugEnd1Light.setColor(new Color3f(newBugLightI,newBugLightI,newBugLightI));
		       bugEnd2Light.setColor(new Color3f(newBugLightI,newBugLightI,newBugLightI));
		    }
	    }
	    
	    if (source == bugLightI) {
	    	if (!source.getValueIsAdjusting()) {
		       float newBugLightI = (float)(source.getValue())/100;
		       bugEnd1Light.setColor(new Color3f(newBugLightI,newBugLightI,newBugLightI));
		       bugEnd2Light.setColor(new Color3f(newBugLightI,newBugLightI,newBugLightI));
		    }
	    }
	}
	
	
	public static void allBranchGroupCapabilities(BranchGroup bg) {
		bg.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		bg.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
		bg.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
		bg.setCapability(BranchGroup.ALLOW_DETACH);
	}
	
	private static void talkln (String info) {
		System.out.println(info);
	}
	
	private static void talk (String info) {
		System.out.println(info);
	}
	
	public void loadParamsFromFile () {
		String loadFromName = null;
		loadFromName = FileOps.getFileToLoad("Load parameter settings from ...", this);
		if (loadFromName == null) {
			talkln ("Parameter settings not loaded.");
		} else {
			File loadFromFile = new File(loadFromName);
			FileOps.loadParamConfig(loadFromFile, true);
		}
	}
	
	public void saveParamsToFile () {
		String saveToName = null;
		saveToName = FileOps.getFileToSave("Save parameter settings to ...", this);
		if (saveToName == null) {
			talkln ("Parameter settings not saved.");
		} else {
			File saveToFile = new File(saveToName);
			FileOps.writeParamConfig(saveToFile);
		}
	}
	
	public void prepForJPEGWriting () {
		Env.toFilePath = null;
		Env.toFilePath = FileOps.getDirectoryToSave("Create folder for JPEG files ...", this);
		if (Env.toFilePath == null) { 
			Env.toFile = false;
			writeFileMenuItem.setState(false);
		} else {
			int lastbitIndex = Env.toFilePath.lastIndexOf(File.separator);
			Env.toFileName = Env.toFilePath.substring(lastbitIndex+1);
			theCanvas.capturePath = Env.toFilePath;
			theCanvas.captureName = Env.toFileName;
			Env.toFile = true;
		}
	}
	
	public static void prepForQKWriting () {
		Env.qkFilePath = null;
		Env.qkFilePath = FileOps.getDirectoryToSave("Create folder for QK files ...", null);
		if (Env.qkFilePath == null) { 
			Env.toQKFile = false;
			writeQKFileMenuItem.setState(false);
		} else {
			int lastbitIndex = Env.qkFilePath.lastIndexOf(File.separator);
			Env.outFileName = Env.qkFilePath.substring(lastbitIndex+1);
			Env.toQKFile = true;
		}
	}
	
	public static void prepForQKRender () {
		Env.fromQKFilePath = null;
		Env.fromQKFilePath = FileOps.getDirectoryToLoad("Select folder of QK files to render ...", null);
		if (Env.fromQKFilePath == null) { 
			Env.fromQKFileName = null;
			Env.fromQKFile = false;
		} else {
			int lastbitIndex = Env.fromQKFilePath.lastIndexOf(File.separator);
			Env.fromQKFileName = Env.fromQKFilePath.substring(lastbitIndex+1);
			Env.fromQKFile = true;
		}
	}
	
	public static void prepForQKStateLoad () {
		Env.fromQKStatePath = null;
		Env.fromQKStatePath = FileOps.getFileToLoad("Select QK file state to load ...", null);
	}


	//--------------------------------------------------------------
	//  USER INTERFACE
	//--------------------------------------------------------------


	/**
	 *  Shows the application's frame, making it and its menubar,
	 *  3D canvas, and 3D content visible.
	 */
	public void showFrame( )
	{
		theFrame.setVisible(true);
	}


	/**
	 *  Quits the application.
	 */
	public void quit( )
	{
		System.exit( 0 );
	}



	/**
	 *  Handles checkbox items on a CheckboxMenu.  The Example class has
	 *  none of its own, but subclasses may have some.
	 *
	 *  @param   menu    which CheckboxMenu needs action
	 *  @param   check   which CheckboxMenu item has changed
	 */
	public void checkboxChanged( CheckboxMenu menu, int check )
	{
		// None for us
	}


	/**
	 *  Handles a window closing event notifying the application
	 *  that the user has chosen to close the application without
	 *  selecting the "Exit" menu item.
	 *
	 *  @param   event   a WindowEvent indicating the window is closing
	 */
	public void windowClosing( WindowEvent event )
	{
		quit( );
	}
	public void windowClosed( WindowEvent event )
	{
	}
	public void windowOpened( WindowEvent event )
	{
	}
	public void windowIconified( WindowEvent event )
	{
	}
	public void windowDeiconified( WindowEvent event )
	{
	}
	public void windowActivated( WindowEvent event )
	{
	}
	public void windowDeactivated( WindowEvent event )
	{
	}



}